#include "HShadowMap.h"
#include <cmath>

HShader *HShadowMap::_shader = 0;

/*! \brief ShadowMap Object
 * \param texWidth Must be power of 2
 * \param texHeight Must be power of 2 */
HShadowMap::HShadowMap(int texWidth, int texHeight)
	: _TEXTURE_TYPE(GL_TEXTURE_2D), _TEXTURE_UNIT(GL_TEXTURE1) {

	_texWidth = texWidth;
	_texHeight = texHeight;

	if (_shader == 0) {
		 _shader = new HShader("shadow");
		 // Depth Texture
		 glEnable(_TEXTURE_TYPE);
	}

	activateTex();

	glGenTextures(1, &_texture);
	glBindTexture(_TEXTURE_TYPE, _texture);

	glTexImage2D(_TEXTURE_TYPE, 0, GL_DEPTH_COMPONENT, _texWidth, _texHeight, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

	setupTex();

	// Framebuffer, Renderbuffer
	glGenFramebuffersEXT(1, &_frameBuffer);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _frameBuffer);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, _TEXTURE_TYPE, _texture, 0);
	glDrawBuffer(GL_NONE);
	glReadBuffer(GL_NONE);

	checkFrameBufferStatus();
	glError();

	glBindTexture(_TEXTURE_TYPE, 0);	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

HShadowMap::~HShadowMap() {
	glDeleteFramebuffersEXT(1, &_frameBuffer);
	glDeleteTextures(1, &_texture);
}

void HShadowMap::setViewMatrix(
	GLdouble eyeX, GLdouble eyeY, GLdouble eyeZ,
	GLdouble centerX, GLdouble centerY, GLdouble centerZ,
	GLdouble upX, GLdouble upY, GLdouble upZ) {

	float l = sqrt(upX*upX + upY*upY + upZ*upZ);

	upX /= l;
	upY /= l;
	upZ /= l;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(eyeX, eyeY, eyeZ, centerX, centerY, centerZ, upX, upY, upZ);
	glGetFloatv(GL_MODELVIEW_MATRIX, _viewMatrix);
	glPopMatrix();

	_viewDirection[0] = centerX - eyeX;
	_viewDirection[1] = centerY - eyeY;
	_viewDirection[2] = centerZ - eyeZ;

	l = sqrt(
		_viewDirection[0]*_viewDirection[0] +
		_viewDirection[1]*_viewDirection[1] +
		_viewDirection[2]*_viewDirection[2]);

	_viewDirection[0] /= l;
	_viewDirection[1] /= l;
	_viewDirection[2] /= l;
	
	_viewUpVector[0] = upX;
	_viewUpVector[1] = upY;
	_viewUpVector[2] = upZ;
}

void HShadowMap::setLightViewMatrix() {
	float upX, upY, upZ;

	float rotation =
		_spotLight->_direction[0] * _viewDirection[0] +
		_spotLight->_direction[1] * _viewDirection[1] +
		_spotLight->_direction[2] * _viewDirection[2];

	if (abs(rotation) >= 0.9999999f) {	// 180 degrees
		upX = _viewUpVector[0];
		upY = _viewUpVector[1];
		upZ = _viewUpVector[2];
	}
	else {
		// cross product
		upX = _spotLight->_direction[1] * _viewDirection[2] - _spotLight->_direction[2] * _viewDirection[1];
		upY = _spotLight->_direction[2] * _viewDirection[0] - _spotLight->_direction[0] * _viewDirection[2];
		upZ	= _spotLight->_direction[0] * _viewDirection[1] - _spotLight->_direction[1] * _viewDirection[0];

		float l = sqrt(upX*upX + upY*upY + upZ*upZ);
		
		upX /= l;
		upY /= l;
		upZ /= l;
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(
		_spotLight->_position[0], _spotLight->_position[1],	_spotLight->_position[2],
		_spotLight->_position[0] + _spotLight->_direction[0],
		_spotLight->_position[1] + _spotLight->_direction[1],
		_spotLight->_position[2] + _spotLight->_direction[2],
		upX, upY, upZ
	);
	glGetFloatv(GL_MODELVIEW_MATRIX, _lightViewMatrix);
	glPopMatrix();

	_upX = upX;
	_upY = upY;
	_upZ = upZ;
}

void HShadowMap::setLightProjectionMatrix(GLdouble fovy, GLdouble zNear, GLdouble zFar) {

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(fovy,  (float)_texWidth / (float)_texHeight, zNear, zFar);
	glGetFloatv(GL_PROJECTION_MATRIX, _lightProjectionMatrix);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void HShadowMap::setTextureMatrixToLightSpaceTransform() {
	static GLfloat shiftScale[16] = {
		0.5f, 0.0f, 0.0f, 0.0f, 
		0.0f, 0.5f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.5f, 0.0f,
		0.5f, 0.5f, 0.5f, 1.0f
	};

	glActiveTexture(GL_TEXTURE0);
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(shiftScale);
	glMultMatrixf(_lightProjectionMatrix);
	glMultMatrixf(_lightViewMatrix);
	
	glActiveTexture(GL_TEXTURE1);
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(_viewMatrix);

	glMatrixMode(GL_MODELVIEW);
}

void HShadowMap::beginRenderDepth() {
	setLightViewMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadMatrixf(_lightViewMatrix);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixf(_lightProjectionMatrix);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _frameBuffer);
	_shader->useShader();
	
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0.0f, 1.0f);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, _texWidth, _texHeight);

	glClear(GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
}

void HShadowMap::endRenderDepth() {
	glDisable(GL_POLYGON_OFFSET_FILL);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glUseProgram(0);
}

void HShadowMap::bindTexture() {
	glActiveTexture(_TEXTURE_UNIT);
	glBindTexture(_TEXTURE_TYPE, _texture);
}

void HShadowMap::unbindTexture() {
	glActiveTexture(_TEXTURE_UNIT);
	glBindTexture(_TEXTURE_TYPE, 0);
}

void HShadowMap::bindTextureToShader(string varName, HShader &shader) {
	shader.setUniformi(varName, _TEXTURE_UNIT - GL_TEXTURE0);
}

void HShadowMap::loadViewMatrix() {
	glLoadMatrixf(_viewMatrix);
}

void HShadowMap::assignLight(HSpotLight *spotLight) {
	_spotLight = spotLight;
	setLightProjectionMatrix(spotLight->_cutoff * 2.0, 0.01f, 100.0f);
}

void HShadowMap::activateTex() {
	// Depth Texture
	glActiveTexture(_TEXTURE_UNIT);
}

void HShadowMap::setupTex() {
	glTexParameteri(_TEXTURE_TYPE, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(_TEXTURE_TYPE, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(_TEXTURE_TYPE, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(_TEXTURE_TYPE, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(_TEXTURE_TYPE, GL_DEPTH_TEXTURE_MODE, GL_INTENSITY);
}