#include "HShadowMapSpot.h"
#include <cmath>

/*! \brief ShadowMap Object
 * \param texWidth Must be power of 2
 * \param texHeight Must be power of 2 */
HShadowMapSpot::HShadowMapSpot(int texWidth, int texHeight)
	: HShadowMap(texWidth, texHeight) {

	activateDepthTex();

	glGenTextures(1, &_textureDepth);
	glBindTexture(_TEXTURE_TYPE_DEPTH, _textureDepth);
	glTexImage2D(_TEXTURE_TYPE_DEPTH, 0, GL_DEPTH_COMPONENT , _texWidth, _texHeight, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, NULL);
	setupDepthTex();
	// Framebuffer, Renderbuffer
	glGenFramebuffersEXT(1, &_frameBufferDepth);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _frameBufferDepth);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, _TEXTURE_TYPE_DEPTH, _textureDepth, 0);
	glDrawBuffer(GL_NONE);
	glReadBuffer(GL_NONE);

	checkFrameBufferStatus();
	glError();

	glBindTexture(_TEXTURE_TYPE_DEPTH, 0);	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

HShadowMapSpot::~HShadowMapSpot() {

}

void HShadowMapSpot::setLightViewMatrix() {
	float upX, upY, upZ;

	float rotation =
		_spotLight->_direction[0] * _viewDirection[0] +
		_spotLight->_direction[1] * _viewDirection[1] +
		_spotLight->_direction[2] * _viewDirection[2];

	if (abs(rotation) >= 0.9999999f) {	// 180 degrees
		upX = _viewUpVector[0];
		upY = _viewUpVector[1];
		upZ = _viewUpVector[2];
	}
	else {
		// cross product
		upX = _spotLight->_direction[1] * _viewDirection[2] - _spotLight->_direction[2] * _viewDirection[1];
		upY = _spotLight->_direction[2] * _viewDirection[0] - _spotLight->_direction[0] * _viewDirection[2];
		upZ	= _spotLight->_direction[0] * _viewDirection[1] - _spotLight->_direction[1] * _viewDirection[0];

		float l = sqrt(upX*upX + upY*upY + upZ*upZ);
		
		upX /= l;
		upY /= l;
		upZ /= l;
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	gluLookAt(
		_spotLight->_position[0], _spotLight->_position[1],	_spotLight->_position[2],
		_spotLight->_position[0] + _spotLight->_direction[0],
		_spotLight->_position[1] + _spotLight->_direction[1],
		_spotLight->_position[2] + _spotLight->_direction[2],
		upX, upY, upZ
	);
	glGetFloatv(GL_MODELVIEW_MATRIX, _lightViewMatrix);
	glPopMatrix();

	_upX = upX;
	_upY = upY;
	_upZ = upZ;
}

void HShadowMapSpot::setTextureMatrixToLightSpaceTransform(GLuint texUnit) {
	static GLfloat shiftScale[16] = {
		0.5f, 0.0f, 0.0f, 0.0f, 
		0.0f, 0.5f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.5f, 0.0f,
		0.5f, 0.5f, 0.5f, 1.0f
	};

	glActiveTexture(texUnit);
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(shiftScale);
	glMultMatrixf(_lightProjectionMatrix);
	glMultMatrixf(_lightViewMatrix);
	
	glMatrixMode(GL_MODELVIEW);
}

void HShadowMapSpot::beginRenderDepth() {
	setLightViewMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadMatrixf(_lightViewMatrix);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadMatrixf(_lightProjectionMatrix);

	activateDepthTex();
	glBindTexture(_TEXTURE_TYPE_DEPTH, 0);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _frameBufferDepth);
	
	_shaderDepth->useShader();
	
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(10.0f, 10.0f);

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, _texWidth, _texHeight);

	glClear(GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
}

void HShadowMapSpot::endRenderDepth() {
	glDisable(GL_POLYGON_OFFSET_FILL);

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glUseProgram(0);
}

void HShadowMapSpot::assignLight(HLight *light) {
	_spotLight = (HSpotLight*)light;
	setLightProjectionMatrix(_spotLight->_cutoff * 2.0, 0.1f, 20.0f);
}